#!/usr/bin/env node
// Basic proxy server implementation.
//
// This works by monitoring a descriptor file specified as the first argument,
// which is written out periodically whenever coronerd has updates.  The
// second argument specifies a directory this program can write to.
//
// The descriptor contents look like this:
// {
//   "console": {
//     "hostname": "127.0.0.1",
//     "type": "http/console",
//     "ssl": false,
//     "service": "5080"
//   },
//   "post": {
//     "hostname": "127.0.0.1",
//     "type": "http/writer",
//     "ssl": false,
//     "service": "5080"
//   },
//   "bind": [
//     {
//       "hostname": "0.0.0.0",
//       "type": "http/console",
//       "ssl": true,
//       "service": "4567"
//     },
//     { ... },
//     { ... }
//   ],
//   "ssl": {
//     "certificate_chain_file": "/etc/coronerd/ssl/chain.crt",
//     "certificate": "/etc/coronerd/ssl/foo.crt",
//     "key": "/etc/coronerd/ssl/sp.backtrace.io.key",
//     "other_flags": ...
//   },
//   "proxy": {
//     "/api/simple": "http://localhost:8081",
//     "/api/saml": "http://localhost:8082",
//     [...]
//   }
// }
//
// The expectation is that every proxy target accepts that only requests for
// their specific subset of the full namespace will be forwarded to them.
// api/post is forwarded to the post listener.  Everything else is forwarded
// to the console listener.
//
// Additionally, to remove the need for any npm modules, this script uses
// only built-in node libraries.
//
// To allow for both a dedicated proxy server, and for integration with a
// system nginx service, separate files are written out for both a top-level
// nginx configuration and a service-specific configuration.  Whenever these
// are updated, the server pid file is used to notify the nginx service.
//

const fs        = require('fs');
const net       = require('net');
const path      = require('path');
const util      = require('util');
const cp        = require('child_process');

var descriptor = null;
var last_update = undefined;
var cur_update = undefined;
var watcher = null;
var proxy_prefix = null;
var nginx_pidfile = null;
var nginx_cfgfile = null;
var nginx_vhfile = null;
var nginx_temp_paths = [];
var logstream = null;
var inited = false;
var startup_ts = null;
const post_endpoints = [
  "post",
  "api/post",
  "minidump/post",
  "api/minidump/post",
  "api/timeseries/submit_datapoints",
];

var nginx_supports_tls_v13 = false;
var nginx_supports_reuseport = false;
var nginx_supports_http2 = false;
var nginx_supports_zone = false;
var nginx_supports_max_conns = false;

/* These are appended to support TLSv1.1 / TLSv1 on the POST port. */
const weakened_cipher_list = ":DHE-RSA-CHACHA20-POLY1305:ECDHE-ECDSA-AES128-SHA256:ECDHE-RSA-AES128-SHA256:ECDHE-ECDSA-AES128-SHA:ECDHE-RSA-AES128-SHA:ECDHE-ECDSA-AES256-SHA384:ECDHE-RSA-AES256-SHA384:ECDHE-ECDSA-AES256-SHA:ECDHE-RSA-AES256-SHA:DHE-RSA-AES128-SHA256:DHE-RSA-AES256-SHA256:AES128-GCM-SHA256:AES256-GCM-SHA384:AES128-SHA256:AES256-SHA256:AES128-SHA:AES256-SHA:DES-CBC3-SHA";

/* Appended to any protocol listed in a server. */
var default_ssl_protocols = "TLSv1.2";
/* Updated to "reuseport" when nginx supports that setting. */
var reuseport_string = "";
/* Updated to "http2" when nginx was built with http2 support. */
var http2_string = "";

/*
 * Recent nginx can limit the number of connections made to a specific
 * upstream. At first, use these very lenient limits to be conservative.
 * We should definitely lower the limit for console connections (where
 * the requests are expected to be fast).
 */
const console_max_conns = 1000;
/*
 * Higher limit for posts: they're expected to be big and sometimes
 * slow, and the handlers do all that I/O with async code.
 */
const post_max_conns = 100000;

/*
 * Taken from https://github.com/substack/semver-compare (MIT licensed
 * by James Halliday).
 */
function version_cmp(a, b) {
    var pa = a.split('.');
    var pb = b.split('.');
    for (var i = 0; i < 3; i++) {
        var na = Number(pa[i]);
        var nb = Number(pb[i]);
        if (na > nb) return 1;
        if (nb > na) return -1;
        if (!isNaN(na) && isNaN(nb)) return 1;
        if (isNaN(na) && !isNaN(nb)) return -1;
    }
    return 0;
};

function log(level, str) {
  if (level === 'debug' && !process.env.BT_PROXY_DEBUG)
    return;
  if (logstream)
    logstream.write(`${Date()} cr_proxy.js[${level}] ${str}\n`);
}

function nginx_print(str, indent) {
  return " ".repeat(indent) + str + "\n";
}

function obj_to_s(obj) {
  return JSON.stringify(obj, null, 4);
}

function notify_nginx() {
  var pid = parseInt(fs.readFileSync(nginx_pidfile));
  log("debug", `Sending SIGHUP to ${pid} ...`);
  process.kill(pid, 'SIGHUP');
}

async function readJSONFile(filepath) {
  let filecontents = "";
  let obj = {};
  try {
    filecontents = await fs.promises.readFile(filepath, { encoding: 'utf8' });
    obj = JSON.parse(filecontents);
  } catch (e) {
    return undefined;
  }
  return obj;
}

/*
 * Convert an object into an nginx config file.  This uses a key-value pair
 * object model, with keys being config items, and values being their values.
 * A config item may have multiple values, in which case the value for the
 * key should be an array of all the values to be printed.
 *
 * Misc notes:
 * - nginx sections are always objects in this representation.
 * - POD numbers may be values, but are treated as strings in translation.
 * - e.g. "location x" is not an array, since the location parameters are
 *   part of the key.
 */
function obj2nginxcfg(obj, indent) {
  var cfg = "";

  if (!indent)
    indent = 0;

  for (key in obj) {
    const val = obj[key];

    if (Array.isArray(val)) {
      val.forEach((elem) => {
        if (typeof(elem) !== 'string' && typeof(elem) !== 'number')
          throw new Error("nginx array values must be POD");
        cfg += nginx_print(key + " " + elem + ";", indent);
      });
      continue;
    }

    if (typeof(val) === 'string' || typeof(val) === 'number') {
      cfg += nginx_print(key + " " + val + ";", indent);
      continue;
    }

    cfg += nginx_print(key + " {", indent);
    cfg += obj2nginxcfg(val, indent + 2);
    cfg += nginx_print("}", indent);
  }
  return cfg;
}

/*
 * Helper that returns a promise which resolves if the port given is
 * available for use.  Any error is treated as a reject, while a listening
 * event is treated as a resolve, once the port closes again.
 */
const port_available = (port) => new Promise((resolve, reject) => {
  const srv = net.createServer()
  .once('error', err => reject(err))
  .once('listening', () => {
    const srv_port = srv.address().port;
    srv.once('close', () => resolve(srv_port)).close();
  })
  .listen(port)
});

/*
 * Helper that returns a promise which resolves with a port number once an
 * available port number is found.  A preferred port may be used to start.
 */
const get_port = (port) => new Promise((resolve, reject) => {
  log("debug", `trying port ${port} ...`);
  return port_available(port).then((port) => {
    log("debug", `resolved with port ${port} ...`);
    resolve(port);
  }).catch((err) => {
    /* Try again and let the kernel give us an usable port. */
    log("debug", `got err ${err}`);
    port_available(0).then((port) => {
      log("debug", `backup resolved with port ${port} ...`);
      resolve(port);
    }).catch((err) => {
      log("debug", `backup err: ${err} ...`);
      reject(err);
    });
  });
});

function stat_equal(st0, st1) {
  return st0.inode === st1.inode &&
    st0.size === st1.size &&
    st0.mtimeMs === st1.mtimeMs &&
    st0.ctimeMs === st1.ctimeMs;
}

function nginx_config_opts() {
  const nginx_path = process.env.NGINX_PATH || "/usr/sbin/nginx";
  let params = [];
  try {
    const result = cp.execSync(`${nginx_path} -V 2>&1`).toString();
    const cfg = result.split("\n").find((s) => {
      return s.startsWith("configure arguments:");
    });
    params = cfg.split("--");
    params.shift();

    return params.reduce((accum, val) => {
      const kv = val.split("=");
      if (kv.length === 2 && kv[0].endsWith("-temp-path")) {
        accum.push(kv[0].replace("http-", "").replace(/-/gi, "_"));
      }
      return accum;
    }, []);
  } catch (e) {
    log("error", `Unable to query nginx vars: ${e}`);
    log("error", e.stdout);
  }
  return [];
}

function nginx_set_feature_flags() {
  const nginx_path = process.env.NGINX_PATH || "/usr/sbin/nginx";
  try {
    const result = cp.execSync(`${nginx_path} -V 2>&1`).toString();
    if (result.search("--with-http_v2_module ") != -1)
      nginx_supports_http2 = true;

    const version = cp.execSync(`${nginx_path} -v 2>&1`).toString().replace(/^.*nginx\//, "");
    if (version_cmp(version, "1.13.0") >= 0)
      nginx_supports_tls_v13 = true;

    if (version_cmp(version, "1.9.1") >= 0)
      nginx_supports_reuseport = true;

    /* The directive appeared in 1.9.0, but had to live somewhere else. */
    if (version_cmp(version, "1.13.3") >= 0)
      nginx_supports_zone = true;

    if (version_cmp(version, "1.11.5") >= 0)
      nginx_supports_max_conns = true;
  } catch (e) {
    log("error", `Unable to query nginx vars: ${e}`);
    log("error", e.stdout);
  }

  if (nginx_supports_tls_v13)
    default_ssl_protocols = "TLSv1.2 TLSv1.3";
  if (nginx_supports_reuseport)
    reuseport_string = "reuseport";
  if (nginx_supports_http2)
    http2_string = "http2";
}

async function stat(fname) {
  try {
    const st = await fs.promises.lstat(fname);
    return st;
  } catch (e) {
  }
  return null;
}

/*
 * Get the _all.json descriptor configuration.  This returns:
 * - null: The descriptor is unchanged since the last check.
 * - object: The descriptor exists and requires updating.
 */
async function get_descriptor_config() {
  /*
   * Determine if an update has actually occurred.  There may be multiple
   * pending calls to this function at once, so exclude duplicates.  If this
   * throws an exception, the timer will try again in a second, in case the
   * file is missing or has a parse error.
   *
   * Note that coronerd will change startup_ts whenever it restarts, so
   * this is treated as a change event.
   */
  const desc = await readJSONFile(descriptor);
  const old_ts = startup_ts;

  cur_update = desc.generation;
  if (startup_ts !== null && startup_ts === desc.startup_ts) {
    if (last_update !== undefined && cur_update === last_update) {
      log("debug", `descriptor appears unchanged (${desc.startup_ts}+${cur_update}) since last check (${startup_ts}+${last_update})`);
      return null;
    }
  } else {
    log("debug", `startup_ts changed (was ${startup_ts} now ${desc.startup_ts})`);
    startup_ts = desc.startup_ts;
  }
  log("debug",
    `descriptor is now ${old_ts}+${cur_update}, was ${startup_ts}+${last_update}`);
  return desc;
}

function check_ssl(server, ssl_cfg, weaken_defaults) {
  if (!ssl_cfg.key) {
    log("error", "No SSL key, SSL listeners will be disabled");
    return;
  }
  if (!ssl_cfg.certificate_chain_file && !ssl_cfg.certificate_file) {
    log("error", "No SSL certificate or chain, SSL listeners will be disabled");
    return;
  }
  server.ssl_certificate = ssl_cfg.certificate_chain_file || ssl_cfg.certificate_file;
  server.ssl_certificate_key = ssl_cfg.key;
  server.ssl_prefer_server_ciphers = ssl_cfg.prefer_server_ciphers ? "on" : "off";

  ["protocols", "ciphers", "dhparam", "ecdh_curve"].forEach((param) => {
    if (ssl_cfg[param])
      server[`ssl_${param}`] = ssl_cfg[param];
  });

  if (!("ssl_protocols" in server)) {
    server["ssl_protocols"] = default_ssl_protocols;
    if (weaken_defaults) {
      server["ssl_protocols"] += " TLSv1.1 TLSv1";
      server["ssl_ciphers"] += weakened_cipher_list;
    }
  }
}

async function update_config() {
  var proxy = null;
  /* Nginx must expect to be run with `-p`, so all paths are relative. */
  var nginx_top = {
    daemon: 'off',
    pid: nginx_pidfile,
    worker_processes: "auto",
    events: {
      worker_connections: 2000,
    },
    http: {
      client_max_body_size: 0,

      proxy_read_timeout: "10m",
      proxy_buffering: "off",
      /* Enable HTTP 1.1 for buffering-free chunked encoded requests. */
      proxy_http_version: "1.1",
      proxy_set_header: [
        "X-Proxied-For $remote_addr:",
        /* coronerd needs this to resolve the universe */
        "Host $host",
        /*
         * Disable keep-alive between local nginx and coronerd: it's
         * useless complexity, and seems to be buggy on our end.
         */
        "Connection close",
      ],
      proxy_headers_hash_bucket_size: 64,

      sendfile: "on",
      server_tokens: "off",

      types_hash_max_size: 2048,
      types_hash_bucket_size: 512,

      server_names_hash_bucket_size: 64,
      server_names_hash_max_size: 512,

      keepalive_timeout: 65,
      keepalive_requests: 100,
      client_body_timeout: "60s",
      client_body_buffer_size: "128k",
      send_timeout: "60s",
      lingering_timeout: "5s",
      tcp_nodelay: "on",

      /* Cache a couple thousand handshakes' data for up to 1h. */
      ssl_session_cache: "shared:SSL:10m",
      ssl_session_timeout: "1h",

      gzip: "on",
      gzip_comp_level: 1,
      gzip_disable: "msie6",
      gzip_min_length: 20,
      gzip_http_version: "1.1",
      gzip_proxied: "off",
      gzip_vary: "off",

      include: `${proxy_prefix}/_proxy.nginx.conf.http.d/*`,
      access_log: "off",
      "upstream console": {
      },
      "upstream post": {
      },

      server: {
        listen: [],
        error_page: "500 502 503 504 /status/error.html",
        "location /status": {
          ssi: "on",
          auth_basic: "off",
          root: `${proxy_prefix}/www`,
        },
        /* Set default to forward to the internal console port. */
        "location /": {
          proxy_pass: "http://console",
        },
        /* Add a localhost-only endpoint for stats. */
        "location = /status/nginx": {
          stub_status: "on",

          access_log: "off",
          allow: "127.0.0.1",
          deny: "all",
        },
      },
      "server ": {  /* Hack a second server block this way. */
        listen: [],
        error_page: "500 502 503 504 /status/error.html",
        "location /status": {
          ssi: "on",
          auth_basic: "off",
          root: `${proxy_prefix}/www`,
        },
        "location /": {
          /* coronerd will dispatch between simple POST and multipart in httpd.c */
          proxy_pass: "http://post/api/post",
          proxy_request_buffering: "off",
          /* Redirect unhandled GET to the "this is the submission port" page. */
          "if ($request_method = GET)" : {
            "rewrite ^": "/status/submission_port.html last",
          }
        },
      },
    },
    include: `${proxy_prefix}/_proxy.nginx.conf.d/*`,
  };
  var new_cfg = null;

  if (nginx_supports_zone) {
    nginx_top.http["upstream console"]["zone"] = "console 1m";
    nginx_top.http["upstream post"]["zone"] = "post 1m";
  }
  /* Shortcut var as most stuff goes in this section... */
  var server = nginx_top.http.server;
  var post_server = nginx_top.http["server "];

  /* attempt to override all paths that might be needed. */
  nginx_temp_paths.forEach((tmppath) => {
    nginx_top.http[tmppath] = tmppath.replace("_path", "");
  });

  log("debug", "processing configuration...");
  cur_update = undefined;

  try {
    new_cfg = await get_descriptor_config();
    if (new_cfg === null) {
      return;
    }
  } catch (e) {
    log("error", `failed to process descriptor: ${e}`);
    new_cfg = undefined;
  }

  if (!new_cfg || !new_cfg.proxy || !cur_update) {
    /*
     * If the descriptor is unavailable or does not parse correctly, determine
     * whether it is necessary to create a stub nginx configuration, so
     * that backtrace-nginx can come up.
     */
    try {
      await stat(nginx_cfgfile);
      /* If the call succeeds, the nginx.conf file exists, don't change it. */
      log("debug", "descriptor is missing or incomplete");
      return;
    } catch (st_e) {
      log("info", "creating stub configuration");
      new_cfg = {};
    }
  }

  /* Process the configuration to see if it's actionable. */
  try {
    proxy = new_cfg.proxy;
    ssl = new_cfg.ssl;

    check_ssl(server, new_cfg.ssl, /*weaken_defaults=*/false);
    check_ssl(post_server, new_cfg.ssl, /*weaken_defaults=*/true);

    /*
     * Bind every non-"http/writer" port in the main server block;
     * Coronerd will still enforce per listener type limits on its
     * end.
     *
     * "http/writer" ports are expected to be more exposed, so bind
     * them only to the `post_server` block, which doesn't handle console
     * or bpg routes.
     */
    if (new_cfg.bind) {
      new_cfg.bind.forEach((bindobj) => {
        var target = server.listen;
        if (bindobj.type == "http/writer")
          target = post_server.listen;
        if (bindobj.ssl) {
          if (server.ssl_certificate_key)
            target.push(`${bindobj.service} ssl ${http2_string} ${reuseport_string}`);
        } else {
          target.push(`${bindobj.service} ${reuseport_string}`);
        }
      });
    } else {
      const actual_port = await get_port(443);
      server.listen.push(`${actual_port} ssl ${http2_string} ${reuseport_string}`);
    }

    let conssvc = 80;
    if (new_cfg.console && new_cfg.console.service)
      conssvc = new_cfg.console.service;
    let postsvc = 6097;
    if (new_cfg.post && new_cfg.post.service)
      postsvc = new_cfg.post.service;

    nginx_top.http["upstream console"].server = `127.0.0.1:${conssvc} `;
    nginx_top.http["upstream post"].server  = `127.0.0.1:${postsvc} `;
    if (nginx_supports_max_conns) {
      nginx_top.http["upstream console"].server += ` max_conns=${console_max_conns}`;
      nginx_top.http["upstream post"].server += ` max_conns=${post_max_conns}`;
    }

    /*
     * Forward all incarnations of api/post to the internal post port.
     * We do this for both the main server block and the post server:
     * some customers want the option to submit objects on 443, the same
     * port used for other requests.
     */
    post_endpoints.forEach((ep) => {
      post_server[`location /${ep}`] = {
        proxy_pass: "http://post",
        proxy_request_buffering: "off",
      };
      server[`location /${ep}`] = {
        proxy_pass: "http://post",
        proxy_request_buffering: "off",
      };
    });

    if (proxy) {
      /* Finally, forward all service endpoints. */
      for (key in proxy) {
        server["location " + key] = {
          proxy_pass: proxy[key],
        };
      }
    } else {
      log("debug", "empty descriptor, stubbing");
    }

    /* Write out updated configuration and mv into place. */
    const nginx_vhtmp = nginx_vhfile + ".tmp";
    const nginx_cftmp = nginx_cfgfile + ".tmp";
    await fs.promises.writeFile(nginx_vhtmp, obj2nginxcfg(nginx_top.http));
    await fs.promises.writeFile(nginx_cftmp, obj2nginxcfg(nginx_top));
    await fs.promises.rename(nginx_vhtmp, nginx_vhfile);
    await fs.promises.rename(nginx_cftmp, nginx_cfgfile);

    notify_nginx();
  } catch (e) {
    log("error", `update failed: ${e}`);
    return;
  }

  log("info", `${nginx_cfgfile} (ts ${startup_ts} gen ${cur_update}) updated to: ${obj_to_s(nginx_top)}`);
  last_update = cur_update;
  return;
}

function logpath() {
  if (process.env.BT_PROXY_LOGPATH)
    return process.env.BT_PROXY_LOGPATH;
  if (process.env.BT_PROXY_LOGNAME)
    return `${proxy_prefix}/${process.env.BT_PROXY_LOGNAME}`;
  return null;
}

async function getVarPrefix() {
  const crconf = process.env.CORONERD_CONF || '/etc/coronerd/coronerd.conf';
  const crjson = await readJSONFile(crconf);
  if (crjson && crjson.varprefix) {
    return crjson.varprefix;
  }
  return '/var/run/coronerd';
}

async function do_mkdir(dirpath) {
  const parent = path.dirname(dirpath);
  if (parent !== dirpath) {
    await do_mkdir(parent);
  }
  try {
    await fs.promises.mkdir(dirpath);
  } catch (e) {
    if (e.code !== 'EEXIST')
      throw e;
  }
}

async function mkdirp(dirpath) {
  const normdirpath = path.normalize(dirpath);
  return do_mkdir(normdirpath);
}

async function delay(timeout_ms) {
  return new Promise(resolve => setTimeout(resolve, timeout_ms));
}

async function cp_dir(src_dir, dest_dir) {
  if (fs.existsSync(src_dir) === false)
    return;
  const dirents = await fs.promises.readdir(src_dir);
  return Promise.all(dirents.map((de) => {
    return fs.promises.copyFile(`${src_dir}/${de}`, `${dest_dir}/${de}`);
  }));
}

async function main(argv) {
  /*
   * Process command line arguments.
   */
  if (argv[0] === '-h' || argv[0] === '--help') {
    console.log(`Usage: ${process.argv[1]} <descriptor> <var prefix> <nginx.pid>`);
    process.exit(1);
  }

  /* Initialize context by updating from the descriptor. */
  descriptor = argv[0] || process.env.BT_PROXY_DESCRIPTOR;
  proxy_prefix = argv[1] || process.env.BT_PROXY_PREFIX;
  nginx_pidfile = argv[2] || process.env.BT_PROXY_NGINX_PIDFILE;
  const varprefix = await getVarPrefix();

  if (!descriptor)
    descriptor = `${varprefix}/svclayer/_all.json`;
  if (!proxy_prefix)
    proxy_prefix = `${varprefix}/backtrace-proxy`;
  if (!nginx_pidfile)
    nginx_pidfile = `${proxy_prefix}/_proxy.nginx.pid`;
  nginx_cfgfile = `${proxy_prefix}/nginx.conf`;
  nginx_vhfile = `${proxy_prefix}/nginx.vhost.conf`;
  nginx_temp_paths = nginx_config_opts();
  nginx_set_feature_flags()

  await mkdirp(`${proxy_prefix}/www/status`);
  await cp_dir("/opt/backtrace/coronerd/status", `${proxy_prefix}/www/status`);

  process.chdir(proxy_prefix);
  const logfilepath = logpath();
  if (logfilepath) {
    logstream = fs.openSync(logfilepath, 'a');
    logstream.write("\n----\n");
  } else {
    logstream = process.stdout;
  }

  log('info', `Started, will monitor ${descriptor}, pidfile ${nginx_pidfile}`);
  log('debug', `nginx temp paths: ${JSON.stringify(nginx_temp_paths, null, 4)}`);

  for (;;) {
    await update_config();
    await delay(1000);
  }
}

main(process.argv.slice(2));

// vim:ts=2:sw=2:et
